/*
FILE:     FileObj_Cif_Util.C
*/
/*
VERSION:  10.000
*/
/*
DATE:     6/6/2017
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2017 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               RCSB PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <stdexcept>

#include "CategoryMapping.h"
#include "CifCoordRead.h"
#include "CompositeIndex.h"
#include "FileObj.h"
#include "SeqCodeUtil.h"
#include "SpaceGroup.h"
#include "utillib.h"

#define N_SWITCH_CATEGORY     4
#define N_NAME_CATEGORY       3

bool FileObj::parse_mmcif_file()
{
       if (_CifObj) { delete _CifObj; }
       std::string msg;
       _CifObj = get_fobj(msg, _input_filename, _caseSensitive);
       if (!_CifObj) {
            std::string error = "Read file '" + _input_filename + "' failed";
            if (!msg.empty()) {
                 error += ": ";
                 if (msg.substr(0, 8) == "ERROR - ")
                      error += msg.substr(8);
                 else error += msg;
            } else error += ".";
            throw std::out_of_range(error);
            return false;
       }

       _firstBlockName = _CifObj->GetFirstBlockName();
       _StructureId = _firstBlockName;

       return true;
}

bool FileObj::read_mmcif_file(const bool& coord_and_seq_only, const bool& check_format_flag)
{
       if (!read_mmcif_metadata(coord_and_seq_only)) return false;
       read_mmcif_coordinate(check_format_flag);

       return true;
}

void FileObj::read_mmcif_file_without_parsing(const bool& coord_and_seq_only)
{
       _read_mmcif_metadata(coord_and_seq_only);
       read_mmcif_coordinate();
}

bool FileObj::read_mmcif_metadata(const bool& coord_and_seq_only)
{
       if (!parse_mmcif_file()) return false;

       _read_mmcif_metadata(coord_and_seq_only);

       return true;
}

void FileObj::read_mmcif_coordinate(const bool& check_format_flag)
{
       CifCoordRead reader;
       reader.setLog(_logIo);
       if (check_format_flag) {
            reader.setMessage(&_error_messages);
            reader.setFormatChecking();
       }
       reader.setRecord(&_pdb_records);
       reader.setMolecule(&_molecules);
       reader.setCifObj(_CifObj);
       reader.Read();
}

void FileObj::write_mmcif_file(const bool& internal_flag, const bool& double_quoting_flag,
                               const bool& start_date_flag)
{
       if (!_CifObj || _output_filename.empty()) return;

       Block& _cifblock = _CifObj->GetBlock(_firstBlockName);

       deleteTable(_cifblock, "pdbx_pdb_compnd");
       deleteTable(_cifblock, "pdbx_pdb_source");
/*
       std::vector<std::string> categories;
       categories.clear();
       categories.push_back("entity_src_nat");
       categories.push_back("entity_src_gen");
       categories.push_back("pdbx_entity_src_syn");

       std::set<std::string> items;
       items.clear();
       items.insert("entity_id");

       for (std::vector<std::string>::const_iterator pos =
            categories.begin(); pos != categories.end(); ++pos) {
            ISTable *t = getTablePtr(_cifblock, *pos);
            if (is_empty_table(t, items)) deleteTable(_cifblock, *pos);
       }
*/

       _reformat_one_letter_code_sequence(_cifblock);
       _switch_high_low_resolution(_cifblock);
       _remove_extra_space_in_name(_cifblock);
       _reorder_refine_ls_shell_reflns_shell(_cifblock, "refine_ls_shell");
    
/*
       if (_DepUI_Flag) {
            std::vector<std::string> categories;
            categories.clear();
            categories.push_back("entity_src_nat");
            categories.push_back("entity_src_gen");
            categories.push_back("pdbx_entity_src_syn");
            for (std::vector<std::string>::const_iterator pos =
                 categories.begin(); pos != categories.end(); ++pos) {
                 ISTable *t = getTablePtr(_cifblock, *pos);
                 if (t) continue;
                 t = _newTablePtr(*pos);
                 t->AddRow();
                 _cifblock.WriteTable(t);
            }
       }
*/

       if (start_date_flag) _add_start_date(_cifblock);

       std::vector<std::string> order_list;
       order_list.clear();

       if (internal_flag) {
            _legacy_vs_v5_conversion(_cifblock, MASTER_TO_INTERNAL);
       } else {
            std::string order_file = _rcsbroot  + "/data/ascii/order_list";
            get_order_list(order_file, _cifblock, order_list);
       }

       if (!_resetFirstBlockName.empty()) {
             reset_entry_id(_cifblock, _resetFirstBlockName);
             _CifObj->RenameBlock(_firstBlockName, _resetFirstBlockName);
       }

       if (double_quoting_flag) _CifObj->SetQuoting(CifFile::eDOUBLE);

       if (!order_list.empty() && _CifObj->GetNumBlocks() == 1)
            _CifObj->Write(_output_filename, order_list);
       else _CifObj->Write(_output_filename);
}

void FileObj::_read_mmcif_metadata(const bool& coord_and_seq_only)
{
       Block& _cifblock = _CifObj->GetBlock(_firstBlockName);

       _legacy_vs_v5_conversion(_cifblock, INTERNAL_TO_MASTER);

       _read_mmcif_metadata_process(_cifblock, coord_and_seq_only);
}

void FileObj::_legacy_vs_v5_conversion(Block& block, const int& status)
{
       std::string mapping_file = _rcsbroot + "/data/binary/master_internal_cif_mapping.odb";
       master_internal_conversion(mapping_file, block, status);
}

void FileObj::_read_mmcif_metadata_process(Block& block, const bool& coord_and_seq_only)
{
       if (!coord_and_seq_only) _getExperimentTypefrom_exptl(block);

       _getSequenceInformation(block);

       if (!coord_and_seq_only) {
            _getCrySymmetryfrom_symmetry_and_cell(block);

            _getAuthorDefinedScaleMatrixfrom_atom_sites(block);

            _update_database_PDB_rev(block);

            _updateCrySymmetry();
       }

       std::string cs;

       ISTable *t = getTablePtr(block, "pdbx_data_processing_status");
       if (t) {
            int rowNo = t->GetNumRows();
            for (int i = 0; i < rowNo; ++i) {
                 get_value_clean_lower(cs, t, i, "status");
                 if (cs != "skip") continue;
                 get_value_clean_lower(cs, t, i, "task_name");
                 if (!cs.empty()) _skip_task_set.insert(cs); 
            }
       }

       t = getTablePtr(block, "struct_conn");
       if (t) {
            std::vector<std::string> items1, items2, data;
            items1.clear();
            items1.push_back("ptnr1_auth_asym_id");
            items1.push_back("ptnr1_auth_comp_id");
            items1.push_back("ptnr1_auth_seq_id");
            items1.push_back("pdbx_ptnr1_PDB_ins_code");
            items1.push_back("ptnr2_auth_asym_id");
            items1.push_back("ptnr2_auth_comp_id");
            items1.push_back("ptnr2_auth_seq_id");
            items1.push_back("pdbx_ptnr2_PDB_ins_code");
            items2.clear();
            items2.push_back("ptnr2_auth_asym_id");
            items2.push_back("ptnr2_auth_comp_id");
            items2.push_back("ptnr2_auth_seq_id");
            items2.push_back("pdbx_ptnr2_PDB_ins_code");
            items2.push_back("ptnr1_auth_asym_id");
            items2.push_back("ptnr1_auth_comp_id");
            items2.push_back("ptnr1_auth_seq_id");
            items2.push_back("pdbx_ptnr1_PDB_ins_code");
            
            int rowNo = t->GetNumRows();
            for (int i = 0; i < rowNo; ++i) {
                 get_value_clean(cs, t, i, "ptnr1_symmetry");
                 if (!cs.empty() && cs != "1_555") continue;
                 get_value_clean(cs, t, i, "ptnr2_symmetry");
                 if (!cs.empty() && cs != "1_555") continue;

                 data.clear();
                 for (std::vector<std::string>::const_iterator
                      pos = items1.begin(); pos != items1.end(); ++pos) {
                      get_value_clean(cs, t, i, *pos);
                      data.push_back(cs);
                 }
                 std::string idx = CompositeIndex::getIndex(data);
                 _link_residue_set.insert(idx);

                 data.clear();
                 for (std::vector<std::string>::const_iterator
                      pos = items2.begin(); pos != items2.end(); ++pos) {
                      get_value_clean(cs, t, i, *pos);
                      data.push_back(cs);
                 }
                 idx = CompositeIndex::getIndex(data);
                 _link_residue_set.insert(idx);
            }
       } 
}

void FileObj::_getExperimentTypefrom_exptl(Block& block)
{
       ISTable *t = getTablePtr(block, "exptl");
       if (t) {
            std::string cs;
            int rowNo = t->GetNumRows();
            for (int i = 0; i < rowNo; ++i) {
                 get_value_clean_upper(cs, t, i, "method");
                 if (cs.empty()) continue;
                 _getExperimentType(cs); 
            }
       }
       if (!_experiment_type) _experiment_type = EXPERIMENT_TYPE_BASIC;
       
}

void FileObj::_getSequenceInformation(Block& block)
{
       _getSequenceInformationfrom_entity_poly(block);

       _getSequenceInformationfrom_pdbx_poly_seq_scheme(block);

       _getSequenceType();
}

void FileObj::_getSequenceInformationfrom_pdbx_poly_seq_scheme(Block& block)
{
       ISTable *t = getTablePtr(block, "pdbx_poly_seq_scheme");
       if (!t) return;

       if (!t->IsColumnPresent("mon_id") || !t->IsColumnPresent("seq_id") ||
           !t->IsColumnPresent("pdb_strand_id") || !t->IsColumnPresent("pdb_mon_id") ||
           !t->IsColumnPresent("pdb_seq_num") || !t->IsColumnPresent("pdb_ins_code"))
            return;

       std::set<std::string> FirstIndex;
       FirstIndex.clear();

       _Seq_Scheme_Mapping.clear();

       std::string cs, cs1;
       std::vector<std::string> data, items;
       std::vector<std::vector<std::string> > t_vectors;

       items.clear();
       items.push_back("pdb_strand_id");
       items.push_back("mon_id");
       items.push_back("pdb_mon_id");
       items.push_back("pdb_seq_num");
       items.push_back("pdb_ins_code");
       items.push_back("entity_id");

       int rowNo = t->GetNumRows();
       for (int i = 0; i < rowNo; ++i) {
            get_value_clean_lower(cs, t, i, "hetero");
            if (cs == "y") {
                 get_value(cs1, t, i, "seq_id");
                 cs1 += "_" + cs;
                 if (FirstIndex.find(cs1) != FirstIndex.end()) continue;
                 FirstIndex.insert(cs1);
            }

            data.clear();
            for (unsigned int j = 0; j < items.size(); ++j) {
                 get_value_clean(cs, t, i, items[j]);
                 data.push_back(cs);
            }

            std::map<std::string, std::vector<std::vector<std::string> > >::iterator
                mpos = _Seq_Scheme_Mapping.find(data[0]);
            if (mpos == _Seq_Scheme_Mapping.end()) {
                 t_vectors.clear();
                 t_vectors.push_back(data);
                 _Seq_Scheme_Mapping.insert(std::make_pair(data[0], t_vectors));
                 FirstIndex.clear();
            } else mpos->second.push_back(data);
       }
}

void FileObj::_getSequenceInformationfrom_entity_poly(Block& block)
{
       ISTable *t = getTablePtr(block, "entity_poly");
       if (!t) {
            // deleteTable(block, "pdbx_poly_seq_scheme");
            return;
       }

       // bool found = false;
       std::string cs;
       int rowNo = t->GetNumRows();
/*
       for (int i = 0; i < rowNo; ++i) {
            get_value_clean_lower(cs, t, i, "pdbx_sequence_evidence_code");
            if (cs == "derived from coordinates") {
                 found = true;
                 break;
            }
       }
       if (found) {
            deleteTable(block, "pdbx_poly_seq_scheme");
            deleteTable(block, "entity_poly");
            return;
       }
*/

       if (/* !t->IsColumnPresent("type") || */ !t->IsColumnPresent("pdbx_strand_id")) return;
       if (!t->IsColumnPresent("pdbx_seq_one_letter_code") &&
           !t->IsColumnPresent("pdbx_seq_three_letter_code")) return;

       std::string entity_id, cs1, cs2, PolyType, EvidenceCode;
       std::vector<std::string> data;
       SEQ seq;

       _Seqs.clear();
       for (int i = 0; i < rowNo; ++i) {
            get_value_clean(entity_id, t, i, "entity_id");
            get_value_clean(PolyType, t, i, "type");
            get_value_clean(EvidenceCode, t, i, "pdbx_sequence_evidence_code");
            if (EvidenceCode.empty()) EvidenceCode = "depositor provided";

            get_value_clean_upper(cs, t, i, "pdbx_seq_one_letter_code");
            get_value_clean_upper(cs1, t, i, "pdbx_N_terminal_seq_one_letter_code");
            get_value_clean_upper(cs2, t, i, "pdbx_C_terminal_seq_one_letter_code");
            std::string sequence = cs1 + cs + cs2;

            data.clear();
            if (!sequence.empty()) {
/*
                 if (PolyType.empty()) {
                      if (sequence.find("E") != std::string::npos ||
                          sequence.find("F") != std::string::npos ||
                          sequence.find("K") != std::string::npos ||
                          sequence.find("P") != std::string::npos ||
                          sequence.find("Q") != std::string::npos ||
                          sequence.find("R") != std::string::npos ||
                          sequence.find("V") != std::string::npos ||
                          sequence.find("Y") != std::string::npos)
                           PolyType = "polypeptide(L)";
                      else if (sequence.size() > 40) 
                           PolyType = "polyribonucleotide";
                      else PolyType = "polydeoxyribonucleotide";
                 }
*/
                 SeqCodeUtil::GetThreeLetterCodeSeq(sequence, PolyType, data);
            } else if (t->IsColumnPresent("pdbx_seq_three_letter_code")) {
                 get_value_upper(sequence, t, i, "pdbx_seq_three_letter_code");
                 if (!sequence.empty()) get_wordarray(data, sequence, " \t\n");
            }
            if (data.empty()) continue;

            _clearSEQ(seq);
            seq.entity_id = entity_id;
            seq.res = data;
            seq.PolyType = PolyType;
            seq.EvidenceCode = EvidenceCode;
            get_value(cs, t, i, "pdbx_strand_id");
            get_wordarray(data, cs, ", \t\n");
            for (std::vector<std::string>::const_iterator
                 pos = data.begin(); pos != data.end(); ++pos) {
                 seq.ChainId = *pos;
                 seq.PDB_ChainId = *pos;
                 seq.order = (int) _Seqs.size();
                 _Seqs.insert(std::make_pair(*pos, seq));
            }
       }
}

void FileObj::_getCrySymmetryfrom_symmetry_and_cell(Block& block)
{
       std::string cs;

       ISTable *t = getTablePtr(block, "cell");
       if (t) {
            get_value(cs, t, 0, "length_a");    if (!cs.empty()) _cell.setA(cs);
            get_value(cs, t, 0, "length_b");    if (!cs.empty()) _cell.setB(cs);
            get_value(cs, t, 0, "length_c");    if (!cs.empty()) _cell.setC(cs);
            get_value(cs, t, 0, "angle_alpha"); if (!cs.empty()) _cell.setAlpha(cs);
            get_value(cs, t, 0, "angle_beta");  if (!cs.empty()) _cell.setBeta(cs);
            get_value(cs, t, 0, "angle_gamma"); if (!cs.empty()) _cell.setGamma(cs);
       }

       if (_update_symmetry_flag)
            t = _getTablePtr(block, "symmetry");
       else t = getTablePtr(block, "symmetry");
       if (t) {
            get_value(cs, t, 0, "space_group_name_H-M");
            if (cs.empty()) get_value(cs, t, 0, "pdbx_full_space_group_name_H-M");
            if (!cs.empty()) {
                 std::string number;
                 SpaceGroup::CheckSpaceGroup(cs, number);
                 if (_update_symmetry_flag) {
                      t->UpdateCell(0, "space_group_name_H-M", cs);
                      if (!number.empty()) t->UpdateCell(0, "Int_Tables_number", number);
                      block.WriteTable(t);
                 }
                 _cell.setSpaceGroup(cs);
            }
       }
}

void FileObj::_getAuthorDefinedScaleMatrixfrom_atom_sites(Block& block)
{
       const char *_matrix_items[3][3] = {
            { "fract_transf_matrix[1][1]", "fract_transf_matrix[1][2]", "fract_transf_matrix[1][3]" },
            { "fract_transf_matrix[2][1]", "fract_transf_matrix[2][2]", "fract_transf_matrix[2][3]" },
            { "fract_transf_matrix[3][1]", "fract_transf_matrix[3][2]", "fract_transf_matrix[3][3]" }
       };
       const char *_vector_items[3] = {
               "fract_transf_vector[1]", "fract_transf_vector[2]", "fract_transf_vector[3]"
       };

       _exist_scale_matrix = false;

       ISTable *t = getTablePtr(block, "atom_sites");
       if (!t) return;

       std::string cs;
       for (int i = 0; i < 3; ++i) {
            for (int j = 0; j < 3; ++j) {
                 get_value_clean(cs, t, 0, _matrix_items[i][j]);
                 if (cs.empty()) return;
                 _scale[i][j] = atof(cs.c_str());
            }
            get_value_clean(cs, t, 0, _vector_items[i]);
            if (cs.empty()) return;
            _vec[i] = atof(cs.c_str());
       }
       _exist_scale_matrix = true;
}

void FileObj::_update_database_PDB_rev(Block& block)
{
       ISTable *t = getTablePtr(block, "database_PDB_rev");
       if (!t) {
            t = getTablePtr(block, "pdbx_database_status");
            if (!t) return;

            std::string cs;
            get_value_clean(cs, t, 0, "recvd_initial_deposition_date");
            if (cs.empty()) return;

            t = _newTablePtr("database_PDB_rev");
            t->AddRow();
            t->UpdateCell(0, "date_original", cs);
            block.WriteTable(t);

            return;
       }

       std::vector<std::string> items, data;
       items.clear();
       items.push_back("pdbx_record_revised_1");
       items.push_back("pdbx_record_revised_2");
       items.push_back("pdbx_record_revised_3");
       items.push_back("pdbx_record_revised_4");

       std::map<int, std::vector<std::string> > rev_records;
       rev_records.clear();

       std::set<std::string> index_set;
       std::string cs;
      
       int rowNo = t->GetNumRows();
       for (int i = 0; i < rowNo; ++i) {
            index_set.clear();
            data.clear();
            for (std::vector<std::string>::const_iterator
                 pos = items.begin(); pos != items.end(); ++pos) {
                 get_value_clean(cs, t, i, *pos);
                 if (cs.empty()) continue;
                 if (index_set.find(cs) != index_set.end()) continue;
                 index_set.insert(cs);
                 data.push_back(cs); 
            }
            if (data.empty()) continue;

            get_value_clean(cs, t, i, "num");
            if (cs.empty()) continue;

            rev_records.insert(std::make_pair(atoi(cs.c_str()), data));
       }

       for (std::vector<std::string>::const_iterator
            pos = items.begin(); pos != items.end(); ++pos) {
            if (t->IsColumnPresent(*pos)) t->DeleteColumn(*pos);
       }
       block.WriteTable(t);

       if (rev_records.empty()) return;

       t = getTablePtr(block, "database_PDB_rev_record");
       if (t) {
            std::string cs1;
            index_set.clear();
            rowNo = t->GetNumRows();
            for (int i = 0; i < rowNo; ++i) {
                 get_value_clean(cs, t, i, "rev_num");
                 get_value_clean(cs1, t, i, "type");
                 if (index_set.find(cs + "_" + cs1) != index_set.end()) continue;

                 index_set.insert(cs + "_" + cs1);
                 int num = atoi(cs.c_str());
                 std::map<int, std::vector<std::string> >::iterator
                     mpos = rev_records.find(num);
                 if (mpos != rev_records.end())
                      mpos->second.push_back(cs1);
                 else {
                      data.clear();
                      data.push_back(cs1);
                      rev_records.insert(std::make_pair(num, data));
                 }
            }
       }

       t = _newTablePtr("database_PDB_rev_record");
       int row = 0;
       for (std::map<int, std::vector<std::string> >::const_iterator
            mpos = rev_records.begin(); mpos != rev_records.end(); ++mpos) {
            cs = String::IntToString(mpos->first);
            for (std::vector<std::string>::const_iterator pos =
                 mpos->second.begin(); pos != mpos->second.end(); ++pos) {
                 t->AddRow();
                 t->UpdateCell(row, "rev_num", cs);
                 t->UpdateCell(row, "type", *pos);
                 row++;
            }
       }
       block.WriteTable(t);
}

void FileObj::_reformat_one_letter_code_sequence(Block& block)
{
       std::map<std::string, std::set<std::string> > seq_categories;
       seq_categories.clear();

       std::set<std::string> items;

       items.clear();
       items.insert("pdbx_seq_one_letter_code");
       seq_categories.insert(std::make_pair("struct_ref", items));

       items.clear();
       items.insert("db_seq_one_letter_code");
       seq_categories.insert(std::make_pair("pdbx_struct_ref_seq_depositor_info", items));

       items.clear();
       items.insert("one_letter_code");
       items.insert("one_letter_code_mod");
       seq_categories.insert(std::make_pair("pdbx_seq_map_depositor_info", items));

       std::vector<std::string> data;
       std::string cs;

       for (std::map<std::string, std::set<std::string> >::const_iterator
            mpos = seq_categories.begin(); mpos != seq_categories.end(); ++mpos) {
            ISTable *t = getTablePtr(block, mpos->first);
            if (!t) continue;

            int rowNo = t->GetNumRows();
            for (int i = 0; i < rowNo; ++i) {
                 for (std::set<std::string>::const_iterator spos =
                      mpos->second.begin(); spos != mpos->second.end(); ++spos) {
                      get_value_clean(cs, t, i, *spos);
                      if (cs.empty()) continue;
                      SeqCodeUtil::GetOneLetterCodeSeqArray(cs, data);
                      cs = SeqCodeUtil::GetOneLetterCodeSeq(data, 80);
                      t->UpdateCell(i, *spos, cs);
                 }
            }
            block.WriteTable(t);
       }
}

void FileObj::_switch_high_low_resolution(Block& block)
{
       const char *_switch_categories[N_SWITCH_CATEGORY][3] = {
            { "refine",          "ls_d_res_high",     "ls_d_res_low"     },
            { "reflns",          "d_resolution_high", "d_resolution_low" },
            { "refine_ls_shell", "d_res_high",        "d_res_low"        },
            { "reflns_shell",    "d_res_high",        "d_res_low"        }
       };

       std::string high, low;

       for (int i = 0; i < N_SWITCH_CATEGORY; ++i) {
            ISTable *t = getTablePtr(block, _switch_categories[i][0]);
            if (!t) continue;

            int rowNo = t->GetNumRows();
            for (int j = 0; j < rowNo; ++j) {
                 get_value_clean(high, t, j, _switch_categories[i][1]);
                 if (high.empty()) continue;
                 get_value_clean(low, t, j, _switch_categories[i][2]);
                 if (low.empty()) continue;

                 if (atof(high.c_str()) > atof(low.c_str())) {
                      t->UpdateCell(j, _switch_categories[i][1], low);
                      t->UpdateCell(j, _switch_categories[i][2], high);
                 }
            }
            block.WriteTable(t);
       }
}

void FileObj::_remove_extra_space_in_name(Block& block)
{
       const char *_name_categories[N_NAME_CATEGORY][2] = {
            { "audit_author",    "name"  },
            { "citation_author", "name"  },
            { "citation_editor", "name"  }
       };

       std::string cs;

       for (int i = 0; i < N_NAME_CATEGORY; ++i) {
            ISTable *t = getTablePtr(block, _name_categories[i][0]);
            if (!t) continue;

            int rowNo = t->GetNumRows();
            for (int j = 0; j < rowNo; ++j) {
                 get_value(cs, t, j, _name_categories[i][1]);
                 if (cs.empty()) continue;

                 String::StripAndCompressWs(cs);
                 replace_string(cs, ". ", ".");
                 t->UpdateCell(j, _name_categories[i][1], cs);
            }
            block.WriteTable(t);
       }
}

void FileObj::_reorder_refine_ls_shell_reflns_shell(Block& block, const std::string& category)
{
       ISTable *t = getTablePtr(block, category);
       if (!t) return;

       if (!t->IsColumnPresent("d_res_high")) return;
 
       int rowNo = t->GetNumRows();
       if (rowNo < 2) return;

       std::vector<std::string> pdbx_refine_id;
       pdbx_refine_id.clear();

       // first key: pdbx_refine_id
       // second key: d_res_high
       std::map<std::string, std::multimap<double, std::vector<std::string> > > values;
       values.clear();

       std::multimap<double, std::vector<std::string> > tmp_map;
       std::vector<std::string> data;
       std::string cs, id;

       const std::vector<std::string>& items = t->GetColumnNames();
       for (int i = 0; i < rowNo; ++i) {
            data.clear();
            for (std::vector<std::string>::const_iterator
                 pos = items.begin(); pos != items.end(); ++pos) {
                 get_value_clean(cs, t, i, *pos);
                 if (cs.empty()) cs = ".";
                 data.push_back(cs);
            }
            get_value_clean(cs, t, i, "d_res_high");
            get_value_clean(id, t, i, "pdbx_refine_id");
            if (id.empty()) id = "NULL";

            std::map<std::string, std::multimap<double, std::vector<std::string> > >::iterator
                mpos = values.find(id);
            if (mpos != values.end())
                 mpos->second.insert(std::make_pair(atof(cs.c_str()), data));
            else {
                 pdbx_refine_id.push_back(id);
                 tmp_map.clear();
                 tmp_map.insert(std::make_pair(atof(cs.c_str()), data));
                 values.insert(std::make_pair(id, tmp_map));
            }
       }

       int i = 0;
       for (std::vector<std::string>::const_iterator
            pos = pdbx_refine_id.begin(); pos != pdbx_refine_id.end(); ++pos) {
            std::map<std::string, std::multimap<double, std::vector<std::string> > >::const_iterator
                mpos = values.find(*pos);
            if (mpos == values.end()) continue;

            for (std::multimap<double, std::vector<std::string> >::const_iterator
                 dpos = mpos->second.begin(); dpos != mpos->second.end(); ++dpos) {
                 for (unsigned int j = 0; j < items.size(); ++j) {
                      t->UpdateCell(i, items[j], dpos->second[j]);
                 }
                 i++;
            }
       }
       block.WriteTable(t);
}

void FileObj::_add_start_date(Block& block)
{
       ISTable *t = _getTablePtr(block, "pdbx_database_status");
       if (!t) return;

       std::string cs;
       get_value_clean(cs, t, 0, "date_begin_processing");
       if (!cs.empty()) return;

       get_value_clean_upper(cs, t, 0, "status_code");
       if (cs != "PROC" && cs != "REPL" && cs != "WAIT") return;

       // find existing start date
       cs.clear();
       ISTable *t1 = getTablePtr(block, "pdbx_database_proc");
       if (t1) {
            int rowNo = t1->GetNumRows(); 
            for (int i = 0; i < rowNo; ++i) {
                 get_value_clean(cs, t1, i, "date_begin_cycle");
                 if (!cs.empty()) break;
            }
       }
       if (cs.empty()) get_date(cs);

       t->UpdateCell(0, "date_begin_processing", cs);
       block.WriteTable(t);
}

ISTable* FileObj::_getTablePtr(Block& block, const std::string& catName)
{
       ISTable *t = getTablePtr(block, catName);
       if (!t) return NULL;

       try {
            const CIF_CATEGORY& category = CategoryMapping::find_category(catName);
            check_missing_item(t, category.itemNames);
       } catch (const std::exception& exc) {}

       return t;
}

ISTable* FileObj::_newTablePtr(const CIF_CATEGORY& cifcategory)
{
       return add_new_table(cifcategory.category, cifcategory.itemNames);
}

ISTable* FileObj::_newTablePtr(const std::string& catName)
{
       try {
            const CIF_CATEGORY& category = CategoryMapping::find_category(catName);
            return _newTablePtr(category);
       } catch (const std::exception& exc) {
            if (_dictUtil.Read()) {
                 const std::vector<std::string>& itemNames = _dictUtil.getItems(catName);
                 if (!itemNames.empty()) return add_new_table(catName, itemNames);
            }
            throw std::out_of_range("Category " + catName + " is unrecognized.\n");
       }
       return NULL;
}
