/*
FILE:     Molecule_Util.C
*/
/*
VERSION:  10.000
*/
/*
DATE:     6/6/2017
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2017 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               RCSB PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "CompositeIndex.h"
#include "Molecule.h"
#include "utillib.h"

void Molecule::GetChainsAndWaters(std::vector<Chain*>& polymer_chains, std::vector<Chain*>&
                                  nonpolymer_chains, std::list<Residue*>& water_lists)
{
       polymer_chains.clear();
       nonpolymer_chains.clear();
       water_lists.clear();

       std::vector<Residue*> residues;
       Chain *chain = GetFirstChain();
       while (chain) {
            if (chain->chain_type() == "ATOMP" || chain->chain_type() == "ATOMN")
                 polymer_chains.push_back(chain);
            else if (chain->GetFirstResidue()->ResName() == "HOH" ||
                     chain->GetFirstResidue()->ResName() == "DOD") {
                 chain->GetFirstResidueList(residues);
                 while (!residues.empty()) {
                      for (std::vector<Residue*>::const_iterator
                           vpos = residues.begin(); vpos != residues.end(); ++vpos) {
                           water_lists.push_back(*vpos);
                      }
                      chain->GetNextResidueList(residues);
                 }
            } else nonpolymer_chains.push_back(chain);

            chain = GetNextChain();
       }
}

void Molecule::GetChains(const std::string& pdb_chainid, std::vector<Chain*>& polymer_chains,
                         std::vector<Chain*>& nonpolymer_chains, std::vector<Chain*>& waters)
{
       polymer_chains.clear();
       nonpolymer_chains.clear();
       waters.clear();
       Chain *chain = GetFirstChain();
       while (chain) {
            if (chain->PDB_ChainID() != pdb_chainid) {
                 chain = GetNextChain();
                 continue;
            }

            if (chain->chain_type() == "ATOMP" || chain->chain_type() == "ATOMN")
                 polymer_chains.push_back(chain);
            else if (chain->GetFirstResidue()->ResName() == "HOH" ||
                     chain->GetFirstResidue()->ResName() == "DOD")
                 waters.push_back(chain);
            else nonpolymer_chains.push_back(chain);

            chain = GetNextChain();
       }
}

void Molecule::GetChains(std::set<std::string>& pdb_chainids, std::map<std::string, std::vector<Chain*> >& polymer_chains,
                         std::map<std::string, std::vector<Chain*> >& other_chains)
{
       pdb_chainids.clear();
       polymer_chains.clear();
       other_chains.clear();

       Chain *chain = GetFirstChain();
       while (chain) {
            pdb_chainids.insert(chain->PDB_ChainID());

            if (chain->chain_type() == "ATOMP" || chain->chain_type() == "ATOMN")
                 _insert_chains(polymer_chains, chain);
            else _insert_chains(other_chains, chain);

            chain = GetNextChain();
       }
}

void Molecule::GetChains(std::vector<std::string>& pdb_chain_ids, std::map<std::string, std::vector<Chain*> >& polymer_chains,
                         std::map<std::string, std::vector<Chain*> >& non_polymer_chains, std::map<std::string, std::vector<Chain*> >& water_chains)
{
       pdb_chain_ids.clear();
       polymer_chains.clear();
       non_polymer_chains.clear();
       water_chains.clear();

       std::set<std::string> chain_id_set;
       chain_id_set.clear();

       Chain *chain = GetFirstChain();
       while (chain) {
            if (chain_id_set.find(chain->PDB_ChainID()) == chain_id_set.end()) {
                 chain_id_set.insert(chain->PDB_ChainID());
                 pdb_chain_ids.push_back(chain->PDB_ChainID());
            }

            if (chain->chain_type() == "ATOMP" || chain->chain_type() == "ATOMN")
                 _insert_chains(polymer_chains, chain);
            else if (chain->GetFirstResidue()->ResName() == "HOH" || chain->GetFirstResidue()->ResName() == "DOD")
                 _insert_chains(water_chains, chain);
            else _insert_chains(non_polymer_chains, chain);

            chain = GetNextChain();
       }
}

void Molecule::GetMolInfo(std::vector<std::string>& pdb_chain_ids, std::map<std::string, Chain*>& polymers, std::map<std::string,
                          std::list<Residue*> >& nonpolymers, std::map<std::string, std::list<Residue*> >& waters)
{
       pdb_chain_ids.clear();
       polymers.clear();
       nonpolymers.clear();
       waters.clear();

       std::set<std::string> chain_id_set;
       chain_id_set.clear();

       std::list<Residue*> r_list;

       std::vector<Residue*> residues;
       Chain *chain = GetFirstChain();
       while (chain) {
            std::string chainID = chain->PDB_ChainID();
            if (chain_id_set.find(chainID) == chain_id_set.end()) {
                 chain_id_set.insert(chainID);
                 pdb_chain_ids.push_back(chainID);
            }

            if (chain->chain_type() == "ATOMP" || chain->chain_type() == "ATOMN")
                 polymers.insert(std::make_pair(chainID, chain));
            else if (chain->GetFirstResidue()->ResName() == "HOH" || chain->GetFirstResidue()->ResName() == "DOD") {
                 chain->GetFirstResidueList(residues);
                 while (!residues.empty()) {
                      for (std::vector<Residue*>::const_iterator vpos = residues.begin(); vpos != residues.end(); ++vpos) {
                           std::map<std::string, std::list<Residue*> >::iterator mpos = waters.find(chainID);
                           if (mpos != waters.end()) mpos->second.push_back(*vpos);
                           else {
                                r_list.clear();
                                r_list.push_back(*vpos);
                                waters.insert(std::make_pair(chainID, r_list));
                           }
                      }
                      chain->GetNextResidueList(residues);
                 }
            } else {
                 chain->GetFirstResidueList(residues);
                 while (!residues.empty()) {
                      for (std::vector<Residue*>::const_iterator vpos = residues.begin(); vpos != residues.end(); ++vpos) {
                           std::map<std::string, std::list<Residue*> >::iterator mpos = nonpolymers.find(chainID);
                           if (mpos != nonpolymers.end()) mpos->second.push_back(*vpos);
                           else {
                                r_list.clear();
                                r_list.push_back(*vpos);
                                nonpolymers.insert(std::make_pair(chainID, r_list));
                           }
                      }
                      chain->GetNextResidueList(residues);
                 }
            }

            chain = GetNextChain();
       }
}

bool Molecule::PrdUpdate(const std::vector<std::vector<std::string> >& mapping,
                         const std::map<std::string, SEQ>& seqs)
{
       bool _successful = true;

       if (mapping.empty()) return _successful;

       // integer is chain index
       std::set<int> prd_involved_chain_set;
       _get_PRD_involved_chains(mapping, prd_involved_chain_set, _successful);

       if (!_successful) return _successful;
       if (prd_involved_chain_set.empty()) return _successful;

       // map key: residue ID (pdbchnid_name_number_inscode)
       // value: residue index
       std::multimap<std::string, int> residue_mapping;

       // map key: atom ID (pdbchnid_name_number_inscode_atomname)
       // value: atom pointer
       std::multimap<std::string, Atom*> found_atoms;

       _get_PRD_involved_residue_and_atoms(prd_involved_chain_set, residue_mapping,
                                           found_atoms, _successful);

       if (!_successful) return _successful;
       if (found_atoms.empty()) return _successful;

       _remove_old_residues_and_chains(residue_mapping, prd_involved_chain_set);
       _add_new_residues_and_chains(mapping, found_atoms, seqs, _successful);

       std::vector<Chain*> tmp_chain;
       tmp_chain.clear();
       for (std::vector<Chain*>::const_iterator
            pos = _chains.begin(); pos != _chains.end(); ++pos) {
            Chain* chain = *pos;
            if (chain == NULL) continue;
            tmp_chain.push_back(chain);
       }
       _chains = tmp_chain;

       InternalOrder();

       return _successful;
}

bool Molecule::PrdUpdateAtom(const std::map<std::string, std::map<std::string,
                             std::string> >& atom_mapping)
{
       bool _successful = true;
       if (atom_mapping.empty()) return _successful;

       std::vector<std::string> data;
       for (std::map<std::string, std::map<std::string,std::string> >::const_iterator
            mpos = atom_mapping.begin(); mpos != atom_mapping.end(); ++mpos) {
            std::multimap<std::string, int>::iterator pos = _pdbIndex.find(mpos->first);
            if (pos == _pdbIndex.end()) {
                 get_wordarray(data, mpos->first, "_");
                 if (data.size() == 3) data.push_back("");
                 std::string error = "Residue (" + data[0] + " " + data[1] + " " 
                                   + data[2] + data[3] + ") does not eixt in Model "
                                   + String::IntToString(_Mol_ID) + ".\n";
                 _logIo->message(error.c_str());
                 _successful = false;
                 continue;
            }
            if (!_residues[pos->second]->update_atom_name(mpos->second, _Mol_ID, true))
                 _successful = false;
       }

       return _successful;
}

bool Molecule::PrdUpdateNumbering(const std::map<std::string, std::string>& num_mapping)
{
       bool _successful = true;
       if (num_mapping.empty()) return _successful;

       std::set<int> chain_index_set;
       chain_index_set.clear();

       std::vector<std::string> data;
       for (std::map<std::string, std::string>::const_iterator
            mpos = num_mapping.begin(); mpos != num_mapping.end(); ++mpos) {
            std::multimap<std::string, int>::iterator pos = _pdbIndex.find(mpos->first);
            if (pos == _pdbIndex.end()) {
                 get_wordarray(data, mpos->first, "_");
                 if (data.size() == 3) data.push_back("");
                 std::string error = "Residue (" + data[0] + " " + data[1] + " " 
                                   + data[2] + data[3] + ") does not eixt in Model "
                                   + String::IntToString(_Mol_ID) + ".\n";
                 _logIo->message(error.c_str());
                 _successful = false;
                 continue;
            }

            int index = pos->second;
            _residues[index]->set_pdb_res_no(mpos->second);
            chain_index_set.insert(_residues[index]->chain_index());
            Atom *atom = _residues[index]->GetFirstAtom();
            while (atom) {
                 atom->set_pdb_resnum(mpos->second);
                 atom = _residues[index]->GetNextAtom();
            }

            _pdbIndex.erase(mpos->first);
            std::string cs = CompositeIndex::getIndex(_residues[index]->pdb_chnid(),
                                     _residues[index]->ResName(),
                                     _residues[index]->pdb_res_no(),
                                     _residues[index]->ins_code());
            _pdbIndex.insert(std::make_pair(cs, index));
       }

       if (_successful && !chain_index_set.empty()) {
            for (std::set<int>::const_iterator
                 spos = chain_index_set.begin(); spos != chain_index_set.end(); ++spos) {
                 Chain *chain = GetIndexChain(*spos);
                 if (!chain) continue;
                 chain->update_number();
            }
       }

       return _successful;
}

bool Molecule::MergePolymer(const int& new_entity_id, const std::string& new_asym_id,
                            const std::string& new_chain_id,
                            const std::vector<std::vector<std::string> >& group,
                            std::set<int>& merged_entityids)
{
       merged_entityids.clear();

       bool _successful = true;
       if (group.empty()) return _successful;

       Chain* chain = new Chain;
       chain->setCCDic(_ccDic);
       chain->setLog(_logIo);
       chain->setMessage(_messageIo);

       chain->set_ChainID(new_asym_id);
       chain->set_PDB_ChainID(new_chain_id);
       chain->set_PUB_ChainID(new_chain_id);
       chain->set_prev_entity_id(String::IntToString(new_entity_id));
       chain->set_index(_chain_index);
       chain->set_order(_chain_index);
       _chain_index++;

       std::set<int> removed_chain_set, removed_residue_chain_set;
       std::set<int>  removed_residue_set;
       removed_chain_set.clear();
       removed_residue_chain_set.clear();
       removed_residue_set.clear();

       std::string chain_type = "";
       for (std::vector<std::vector<std::string> >::const_iterator
            pos = group.begin(); pos != group.end(); ++pos) {
            if (pos->size() == 1) {
                 Chain *chn = GetPolyChain((*pos)[0]);
                 if (!chn) {
                      std::string error = "Chain " + (*pos)[0] + " does not exist in Model "
                                        + String::IntToString(_Mol_ID) + ".\n";
                      _logIo->message(error.c_str());
                      _successful = false;
                      continue;
                 }

                 if (chn->chain_type() == "ATOMP")
                      chain_type = chn->chain_type();
                 else if (chn->chain_type() == "ATOMN" && chain_type != "ATOMP")
                      chain_type = chn->chain_type();
                 else if (chn->chain_type() == "ATOMS" && chain_type != "ATOMP" &&
                          chain_type != "ATOMN")
                      chain_type = chn->chain_type();

                 merged_entityids.insert(chn->int_prev_entity_id());
                 removed_chain_set.insert(chn->index());
                 chain->merge_chain(chn);
            } else if (pos->size() == 4) {
                 Residue *res = find_pdb_residue((*pos)[0], (*pos)[1],
                                                 (*pos)[2], (*pos)[3]);
                 if (!res) {
                      std::string error = "Residue (" + (*pos)[0] + " " + (*pos)[1] + " " 
                                   + (*pos)[2] + (*pos)[3] + ") does not eixt in Model "
                                   + String::IntToString(_Mol_ID) + ".\n";
                      _logIo->message(error.c_str());
                      _successful = false;
                      continue;
                 }

                 std::string type = _ccDic->find_residue_type(res->ResName());
                 if (type == "ATOMP")
                      chain_type = type;
                 else if (type == "ATOMN" && chain_type != "ATOMP")
                      chain_type = type;
                 else if (type == "ATOMS" && chain_type != "ATOMP" &&
                          chain_type != "ATOMN")
                      chain_type = type;

                 removed_residue_chain_set.insert(res->chain_index());
                 removed_residue_set.insert(res->index());
                 chain->merge_residue(res);
            }
       }
       if (chain_type.empty()) chain_type = "ATOMP";

       if (!_successful) {
            delete chain;
            return _successful;
       }

       for (std::set<int>::const_iterator spos = removed_residue_chain_set.begin();
            spos != removed_residue_chain_set.end(); ++spos) {
            Chain *chn = GetIndexChain(*spos);
            if (!chn) continue;
            chn->remove_residue(removed_residue_set);
            if (chn->SeqLen() == 0) removed_chain_set.insert(*spos);
       }

       chain->set_chain_type(chain_type);
       chain->renumbering_polymer_chain();
       chain->update_indices();

       bool first = true;
       std::vector<Chain*> tmp_chains;
       tmp_chains.clear();
       for (std::vector<Chain*>::const_iterator
            pos = _chains.begin(); pos != _chains.end(); ++pos) { 
            if (removed_chain_set.find((*pos)->index()) == removed_chain_set.end()) {
                 tmp_chains.push_back(*pos);
                 continue;
            }
            if (first) tmp_chains.push_back(chain);
            first = false;
            delete *pos;
       }
       _chains = tmp_chains;

       GenerateInternalOrderIndex();

       return _successful;
}

bool Molecule::SplitPolymer(const std::vector<std::vector<std::string> >& splits,
                            const std::vector<std::vector<std::string> >& chain_info)
{
       bool _successful = true;
       if (splits.empty() || chain_info.empty()) return _successful;

       Chain *chain = GetPolyChain(chain_info[0][2]);
       if (!chain) {
            std::string error = "Chain " + chain_info[0][2] + " does not exist in Model "
                              + String::IntToString(_Mol_ID) + ".\n";
            _logIo->message(error.c_str());
            _successful = false;
            return _successful;
       }

       std::vector<Chain*> split_chains;
       chain->get_split_chains(splits, chain_info, _chain_index, split_chains);
       if (split_chains.empty()) {
            _successful = false;
            return _successful;
       }

       std::vector<Chain*> tmp_chains;
       tmp_chains.clear();
       for (std::vector<Chain*>::const_iterator
            pos = _chains.begin(); pos != _chains.end(); ++pos) { 
            if (*pos == chain) {
                 delete *pos;
                 for (std::vector<Chain*>::const_iterator
                      spos = split_chains.begin(); spos != split_chains.end(); ++spos) {
                      tmp_chains.push_back(*spos);
                 }
                 continue;
            }
            tmp_chains.push_back(*pos);
       }
       _chains = tmp_chains;

       GenerateInternalOrderIndex();

       return _successful;
}

void Molecule::_get_PRD_involved_chains(const std::vector<std::vector<std::string> >&
                                        mapping, std::set<int>& chain_set,
                                        bool& _successful)
{
       // get unique chain set
       chain_set.clear();

       // value: pdbchnid_name_number_inscode
       std::set<std::string> unique_residue_set;
       unique_residue_set.clear();

       // Index for mapping array
       // vector[0]: model ID
       // vector[1]: PDB Chain ID
       // vector[2]: Residue Name
       // vector[3]: Residue Number
       // vector[4]: Insertion Code
       // vector[5]: Atom Name
       // vector[6]: New Residue Name
       // vector[7]: New Residue Number
       // vector[8]: New Atom Name
       // vector[9]: New polymer type
       // vector[10]: Component ID
       // vector[11]: Old polymer type
       // vector[12]: Old Asym ID
       // vector[13]: New PDB Chain ID
       // vector[14]: New Asym ID
       for (std::vector<std::vector<std::string> >::const_iterator
            pos = mapping.begin(); pos != mapping.end(); ++pos) {
            std::string cs = (*pos)[1] + "_" + (*pos)[2] + "_"
                           + (*pos)[3] + "_" + (*pos)[4];
            if (unique_residue_set.find(cs) != unique_residue_set.end()) continue;

            unique_residue_set.insert(cs);

            std::multimap<std::string, int>::iterator mpos = _pdbIndex.find(cs);
            if (mpos == _pdbIndex.end()) {
                 std::string error = "Residue (" + (*pos)[1] + " " + (*pos)[2] + " " 
                                   + (*pos)[3] + (*pos)[4] + ") does not eixt in Model "
                                   + String::IntToString(_Mol_ID) + ".\n";
                 _logIo->message(error.c_str());
                 _successful = false;
                 continue;
            }
            chain_set.insert(_residues[mpos->second]->chain_index());
       }
}

void Molecule::_get_PRD_involved_residue_and_atoms(const std::set<int>& chain_set,
                                      std::multimap<std::string, int>& residue_mapping,
                                      std::multimap<std::string, Atom*>& found_atoms,
                                      bool& _successful)
{
       residue_mapping.clear();
       found_atoms.clear();

       std::vector<std::string> data;

       std::vector<Residue*> residue_list;
       for (std::set<int>::const_iterator
            spos = chain_set.begin(); spos != chain_set.end(); ++spos) {
            Chain *chain = GetIndexChain(*spos);
            if (chain == NULL) {
                 _logIo->message("Chain index out of range.\n");
                 _successful = false;
                 continue;
            }

            chain->GetFirstResidueList(residue_list);
            while (!residue_list.empty()) {
                 for (std::vector<Residue*>::const_iterator
                      vpos = residue_list.begin(); vpos != residue_list.end(); ++vpos) {
                      std::string cs = CompositeIndex::getIndex((*vpos)->pdb_chnid(),
                                            (*vpos)->ResName(), (*vpos)->pdb_res_no(),
                                            (*vpos)->ins_code());
                      residue_mapping.insert(std::make_pair(cs, (*vpos)->index()));
                      Atom *atom = (*vpos)->GetFirstAtom();
                      while (atom) {
                           data.clear();
                           data.push_back(atom->pdb_chnid());
                           data.push_back(atom->pdb_resnam());
                           data.push_back(atom->pdb_resnum());
                           data.push_back(atom->ins_code());
                           data.push_back(atom->pdb_atmnam());
                           cs = CompositeIndex::getIndex(data);
                           found_atoms.insert(std::make_pair(cs, atom));

                           atom = (*vpos)->GetNextAtom();
                      }
                 }
                 chain->GetNextResidueList(residue_list);
            }
       }
}

void Molecule::_add_new_residues_and_chains(const std::vector<std::vector<std::string> >&
                                            mapping, std::multimap<std::string, Atom*>&
                                            found_atoms, const std::map<std::string, SEQ>&
                                            seqs, bool& _successful)
{
       std::map<std::string, Residue*> residue_mapping;
       residue_mapping.clear();

       std::map<std::string, int> last_number_mapping;
       last_number_mapping.clear();

       std::list<Residue*> residue_list;
       residue_list.clear();

       // Index for mapping array
       // vector[0]: model ID
       // vector[1]: PDB Chain ID
       // vector[2]: Residue Name
       // vector[3]: Residue Number
       // vector[4]: Insertion Code
       // vector[5]: Atom Name
       // vector[6]: New Residue Name
       // vector[7]: New Residue Number
       // vector[8]: New Atom Name
       // vector[9]: New polymer type
       // vector[10]: Component ID
       // vector[11]: Old polymer type
       // vector[12]: Old Asym ID
       // vector[13]: New PDB Chain ID
       // vector[14]: New Asym ID
       // vector[15]: New Entity ID

       // insert mapped atoms
       for (std::vector<std::vector<std::string> >::const_iterator
            pos = mapping.begin(); pos != mapping.end(); ++pos) {
            std::string cs = (*pos)[1] + "_" + (*pos)[2] + "_" + (*pos)[3] + "_"
                           + (*pos)[4] + "_" + (*pos)[5];
            std::pair<std::multimap<std::string, Atom*>::iterator,
                      std::multimap<std::string, Atom*>::iterator>
                range = found_atoms.equal_range(cs);
            if (range.first == range.second) return;

            for (std::multimap<std::string, Atom*>::iterator
                 apos = range.first; apos != range.second; ++apos) {
                 apos->second->set_chnid((*pos)[14]);
                 apos->second->set_pdb_chnid((*pos)[13]);
                 apos->second->set_restype((*pos)[6]);
                 apos->second->set_pdb_resnam((*pos)[6]);
                 apos->second->set_resnum((*pos)[7]);
                 apos->second->set_pdb_resnum((*pos)[7]);
                 apos->second->set_ins_code("");
                 apos->second->set_atmtype((*pos)[8]);
                 apos->second->set_pdb_atmnam((*pos)[8]);
                 if (!(*pos)[15].empty()) apos->second->setValue((*pos)[15], 23);
                 _insert_a_atom(apos->second, residue_mapping, last_number_mapping,
                                residue_list, false, (*pos)[9]);
            }
            found_atoms.erase(cs);
       }

       // insert leftover atoms
       for (std::multimap<std::string, Atom*>::const_iterator
            apos = found_atoms.begin(); apos != found_atoms.end(); ++apos) {
            _insert_a_atom(apos->second, residue_mapping, last_number_mapping,
                           residue_list, false, "non-polymer");
       }
       residue_mapping.clear();

       std::map<std::string, Chain*> chain_mapping;
       chain_mapping.clear();

       for (std::list<Residue*>::iterator
            pos = residue_list.begin(); pos != residue_list.end(); ++pos) {
            (*pos)->set_alt_loc();
            (*pos)->get_atom_type();
            (*pos)->reorder_atoms();
            insert_a_residue(*pos);

            Chain *chain = NULL;

            Atom *atom = (*pos)->GetFirstAtom();
            std::map<std::string, Chain*>::const_iterator
                mpos = chain_mapping.find(atom->chnid());
            if (mpos == chain_mapping.end()) {
                 chain = new Chain;
                 chain->setCCDic(_ccDic);
                 chain->setLog(_logIo);
                 chain->setMessage(_messageIo);
                 chain->set_ChainID(atom->chnid());
                 chain->set_PDB_ChainID(atom->pdb_chnid());
                 chain->set_PUB_ChainID(atom->pub_chnid());
                 chain->set_prev_entity_id(atom->getValue("label_entity_id"));
                 insert_a_chain(chain);
                 chain_mapping.insert(std::make_pair(atom->chnid(), chain));
            } else chain = mpos->second;

            chain->insert_a_residue(*pos);
       }
       residue_list.clear();

       std::string chain_type;
       std::map<std::string, int> all_types;
       for (std::map<std::string, Chain*>::iterator
            mpos = chain_mapping.begin(); mpos != chain_mapping.end(); ++mpos) {
            chain_type.clear();
            std::map<std::string, SEQ>::const_iterator
                spos = seqs.find(mpos->second->ChainID());
            if (spos != seqs.end()) {
                 mpos->second->seq_alignment(spos->second);
                 chain_type = spos->second.chain_type;
            }
            if (chain_type.empty()) {
                 all_types.clear();
                 bool has_ATOMP = false;
                 int count = 0;
                 Residue *residue = mpos->second->GetFirstResidue();
                 while (residue) {
                      count++;
                      add_type(all_types, residue->chain_type());
                      if (residue->chain_type() == "ATOMP") has_ATOMP = true;
                      residue = mpos->second->GetNextResidue();
                 }
                 chain_type = get_type(all_types);
                 if ((chain_type != "ATOMP") && has_ATOMP && (count > 1) && (count < 11)) chain_type = "ATOMP";
            }
            mpos->second->set_chain_type(chain_type);
            if (chain_type == "ATOMP" || chain_type == "ATOMN")
                 mpos->second->set_has_sequence();
       }
       chain_mapping.clear();
}

void Molecule::_remove_old_residues_and_chains(const std::multimap<std::string, int>&
                                    residue_mapping, const std::set<int>& chain_set)
{
       // removed old residues and indices 
       for (std::multimap<std::string, int>::const_iterator
            mpos = residue_mapping.begin(); mpos != residue_mapping.end(); ++mpos) {
            _residues[mpos->second]->clear();
            delete _residues[mpos->second];
            _residues[mpos->second] = NULL;
            _pdbIndex.erase(mpos->first);
       }

       // removed old chains
       for (std::vector<Chain*>::iterator
            pos = _chains.begin(); pos != _chains.end(); ++pos) {
            if (chain_set.find((*pos)->index()) == chain_set.end()) continue;
            delete *pos;
            *pos = NULL;
       }
}

bool Molecule::UpdateInstance(const std::map<std::string, std::pair<std::string,
                              std::map<std::string, std::string> > >& atom_mapping)
{
       if (atom_mapping.empty()) return true;

       std::vector<std::string> data;

       std::set<std::string> unique_index;
       unique_index.clear();

       // bool _successful = true;

       for (std::map<std::string, std::pair<std::string, std::map<std::string, std::string> > >
            ::const_iterator mpos = atom_mapping.begin(); mpos != atom_mapping.end(); ++mpos) {
            get_wordarray(data, mpos->first, "_");
            std::string pdb_chnid = data[1];
            std::string res_name  = data[2];
            std::string res_num   = data[3];
            std::string ins_code  = "";
            if (data.size() == 5) ins_code = data[4];
            std::string index = CompositeIndex::getIndex(pdb_chnid, res_name, res_num, ins_code);
            if (unique_index.find(index) != unique_index.end()) continue;

            unique_index.insert(index);

            Residue *res = find_pdb_residue(pdb_chnid, res_name, res_num, ins_code);
            if (!res) {
                 std::string error = "Residue (" + pdb_chnid + " " + res_name + " " 
                              + res_num + ins_code + ") does not eixt in Model "
                              + String::IntToString(_Mol_ID) + ".\n";
                 _logIo->message(error.c_str());
                 // _successful = false;
                 continue;
            }

            if (!res->update_atom_mapping(_Mol_ID, mpos->second.first, mpos->second.second)) {
                 // _successful = false;
                 continue;
            }
       
            if (res_name != mpos->second.first) {
                 Chain* chain = GetIndexChain(res->chain_index());
                 if (!chain) {
                      std::string error = "Search chain index "
                               + String::IntToString(res->chain_index()) + " in Model "
                               + String::IntToString(_Mol_ID) + " failed.\n";
                      _logIo->message(error.c_str());
                      // _successful = false;
                      continue;
                 }
                 if (!chain->Update_SeqRes(res_name, res_num, ins_code, mpos->second.first)) {
                      // _successful = false;
                      continue;
                 }
            }
       }

       return true;
}

bool Molecule::update_residue_nomenclature(const std::map<std::string, std::map<std::string, std::string> >& mapping)
{
       // key: chain_index
       // pair.first: residue_index
       // pair.second: new pdb chainid
       std::map<int, std::list<std::pair<int, std::string> > > chain_index_mapping;
       chain_index_mapping.clear();

       std::list<std::pair<int, std::string> > t_vec;

       bool _successful = true;
       std::vector<std::string> data;
       std::string new_chainid;

       for (std::map<std::string, std::map<std::string, std::string> >::const_iterator
            mpos = mapping.begin(); mpos != mapping.end(); ++mpos) {
            get_wordarray_delimit_by_string(data, mpos->first, "_");

            Residue *res = find_pdb_residue(data[0], data[1], data[2], data[3]);
            if (res) {
                 res->update_nomenclature(mpos->second);
                 Chain *chain = GetIndexChain(res->chain_index());
                 if (chain) chain->Update_SeqRes(data[1], data[2], data[3], mpos->second);

                 new_chainid.clear();
                 std::map<std::string, std::string>::const_iterator
                     mmpos = mpos->second.find("chainid");
                 if (mmpos != mpos->second.end()) new_chainid = mmpos->second;

                 std::map<int, std::list<std::pair<int, std::string> > >::iterator
                     cpos = chain_index_mapping.find(res->chain_index());
                 if (cpos != chain_index_mapping.end()) {
                      if (!new_chainid.empty())
                           cpos->second.push_back(std::make_pair(res->index(), new_chainid));
                 } else {
                      t_vec.clear();
                      if (!new_chainid.empty())
                           t_vec.push_back(std::make_pair(res->index(), new_chainid));
                      chain_index_mapping.insert(std::make_pair(res->chain_index(), t_vec));
                 }
                 continue;
            } else if (mpos->second.find("chainid") == mpos->second.end()) {
                 Chain *chain = GetPolyChain(data[0]);
                 if (chain && chain->Update_SeqRes(data[1], data[2], data[3], mpos->second)) {
                      std::map<int, std::list<std::pair<int, std::string> > >::iterator
                          cpos = chain_index_mapping.find(chain->index());
                      if (cpos == chain_index_mapping.end()) {
                           t_vec.clear();
                           chain_index_mapping.insert(std::make_pair(chain->index(), t_vec));
                      }
                      continue;
                 }
            }

            std::string error = "Residue (" + data[0] + " " + data[1] + " " 
                         + data[2] + data[3] + ") does not eixt in Model "
                         + String::IntToString(_Mol_ID) + ".\n";
            _logIo->message(error.c_str());
            _successful = false;
       }

       if (_successful) _successful = _update_chain_nomenclature(chain_index_mapping);

       return _successful;
}

bool Molecule::_update_chain_nomenclature(const std::map<int, std::list<std::pair<int, std::string> > >& mapping)
{
       if (mapping.empty()) return true;

       std::set<int> removed_chains;
       removed_chains.clear();

       std::list<std::list<Residue*> > relocate_residues;
       std::vector<Chain*> chain_list;

       bool reorder_flag = false;

       for (std::map<int, std::list<std::pair<int, std::string> > >::const_iterator
            mpos = mapping.begin(); mpos != mapping.end(); ++mpos) {
            Chain *chn = GetIndexChain(mpos->first);
            if (!chn) {
                 _logIo->message("Get index chain failed.\n");
                 return false;
            }

            chn->update_nomenclature(mpos->second, relocate_residues);
            for (std::list<std::list<Residue*> >::const_iterator
                 vpos = relocate_residues.begin(); vpos != relocate_residues.end(); ++vpos) {
                 if ((*vpos).front()->ResName() == "HOH" || (*vpos).front()->ResName() == "DOD") {
                      GetPdbChains((*vpos).front()->pdb_chnid(), chain_list);
                      bool found = false;
                      for (std::vector<Chain*>::const_iterator
                           cpos = chain_list.begin(); cpos != chain_list.end(); ++cpos) {
                           if ((*cpos)->empty()) continue;
                           Residue* res = (*cpos)->GetFirstResidue();
                           if (!res) continue;
                           if (res->ResName() == "HOH" || res->ResName() == "DOD") {
                                found = true; 
                                (*cpos)->merge_waters(*vpos);
                                break;
                           }
                      }
                      if (found) continue;

                      insert_a_chain(*vpos);
                      reorder_flag = true;
                 }
            }

            if (chn->empty()) removed_chains.insert(chn->index());
       }

       if (!removed_chains.empty()) {
            std::vector<Chain*> tmp_chain;
            tmp_chain.clear();
            for (std::vector<Chain*>::const_iterator
                 pos = _chains.begin(); pos != _chains.end(); ++pos) { 
                 Chain *chn = *pos;
                 if (removed_chains.find(chn->index()) != removed_chains.end())
                      delete chn;
                 else tmp_chain.push_back(chn);
            }
            _chains = tmp_chain;

            GenerateInternalOrderIndex();
       }

       if (reorder_flag) InternalOrder();

       return true;
}

void Molecule::update_numbering(const std::map<std::string, std::map<std::string, std::string> >& mapping)
{
       std::vector<Chain*> polymer_chains, nonpolymer_chains, water_chains;
       std::map<std::string, std::vector<Chain*> > chain_mapping;
       for (std::map<std::string, std::map<std::string, std::string> >::const_iterator
            mpos = mapping.begin(); mpos != mapping.end(); ++mpos) {
            GetChains(mpos->first, polymer_chains, nonpolymer_chains, water_chains);
            chain_mapping.clear();
            if (!polymer_chains.empty()) chain_mapping.insert(std::make_pair("Polymer", polymer_chains));
            if (!nonpolymer_chains.empty()) chain_mapping.insert(std::make_pair("Nonpolymer", nonpolymer_chains));
            if (!water_chains.empty()) chain_mapping.insert(std::make_pair("Water", water_chains));
            for (std::map<std::string, std::string>::const_iterator
                 mmpos = mpos->second.begin(); mmpos != mpos->second.end(); ++mmpos) {
                 std::map<std::string, std::vector<Chain*> >::const_iterator
                     cpos = chain_mapping.find(mmpos->first);
                 if (cpos == chain_mapping.end()) continue;

                 int serial_no = atoi(mmpos->second.c_str()); 
                 for (std::vector<Chain*>::const_iterator vpos =
                      cpos->second.begin(); vpos != cpos->second.end(); ++vpos) {
                      (*vpos)->update_residues_nomenclature(false, true, serial_no, true);
                 }
            }
       }
}

void Molecule::update_numbering(const std::map<std::string, std::vector<std::string> >& mapping)
{
       for (std::map<std::string, std::vector<std::string> >::const_iterator
            mpos = mapping.begin(); mpos != mapping.end(); ++mpos) {
            Chain* chain = GetPolyChain(mpos->first);
            if (!chain) continue;
            chain->update_residues_nomenclature(mpos->second, true);
       }
}

void Molecule::update_chainid(const std::map<std::string, std::string>& mapping)
{
       for (std::vector<Chain*>::iterator pos = _chains.begin(); pos != _chains.end(); ++pos) {
            std::map<std::string, std::string>::const_iterator
                mpos = mapping.find((*pos)->PDB_ChainID());
            if (mpos != mapping.end()) (*pos)->set_PDB_ChainID_to_residues(mpos->second);
       }
}

void Molecule::_insert_chains(std::map<std::string, std::vector<Chain*> >& mapping_chains, Chain* chain)
{
       std::vector<Chain*> t_vec;
       std::map<std::string, std::vector<Chain*> >::iterator mpos = mapping_chains.find(chain->PDB_ChainID());
       if (mpos != mapping_chains.end()) mpos->second.push_back(chain);
       else {
            t_vec.clear();
            t_vec.push_back(chain);
            mapping_chains.insert(std::make_pair(chain->PDB_ChainID(), t_vec));
       }
}
